package org.greenstone.gatherer.file;
/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * <BR><BR>
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * <BR><BR>
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * <BR><BR>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * <BR><BR>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * <BR><BR>
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */
import java.io.*;
import java.lang.*;
import java.nio.charset.*;
import java.util.*;
import javax.swing.*;
import org.greenstone.gatherer.Gatherer;
import org.greenstone.gatherer.gui.FileAssociationDialog;
import org.greenstone.gatherer.msm.MSMUtils;
import org.greenstone.gatherer.util.Utility;
import org.greenstone.gatherer.util.WinRegistry;
import org.w3c.dom.*;
/** Provides a manager for relating a filetype to a command to open that filetype, and an optional icon. Under windows this will attempt to leverage the Windows Registry. If this fails, or under other platforms, the user will be prompted to enter the required details manually. This information will then be stored for any further sessions of the Gatherer. Of course the ability to edit these settings must also be provided. */
public class FileAssociationManager 
    extends HashMap {
	 
    static private final File ASSOC_FILE = new File(Utility.BASE_DIR, "associations.xml");

    public FileAssociationManager() {
	// Attempt to load any previous file associations
	if(ASSOC_FILE.exists()) {
	    load();
	}
    }

    public void destroy() {
	save();
    }

    public String getCommand(File file) {
	String command = null;
	String filename = file.toString();
	// Determine extension
	int index = -1;
	if((index = filename.lastIndexOf(".")) != -1) {
	    String extension = filename.substring(index + 1);
	    command = getCommand(file, extension);
	}
	return command;
    }

    public String getCommand(File file, String extension) {
	// We first check to see if we already know the answer to how to run this program.
	String command = (String) get(extension);
	if(command != null) {
	    command = command + " " + file.getAbsolutePath();
	}
	// If not, and we are on windows, try the registry.
	if(command == null && Utility.isWindows()) {
	    command = WinRegistry.openCommand(file.getAbsolutePath());
	}
	// Failing all that prompt the user to select the program to open this with.
	if(command == null) {
	    FileAssociationDialog dialog = new FileAssociationDialog(this);
	    command = dialog.display(extension);
	    dialog.destroy();
	    dialog = null;
	    if(command != null) {
		put(extension, command);
		command = command + " " + file.getAbsolutePath();
	    }
	}
	///ystem.err.println("Get command: " + extension + " = " + command);
	return command;
    }

    public String getCommandImmediately(String extension) {
	return (String) get(extension);
    }

    public void edit() {
	FileAssociationDialog dialog = new FileAssociationDialog(this);
	dialog.display(null);
	dialog.destroy();
	dialog = null;		  
    }

    private void load() {
	// Read in and parse the document.
	Document document = Utility.parse(ASSOC_FILE, false);
	if(document != null) {
				// Recover all the entry nodes of the association root.
	    for(Node entry_n = document.getDocumentElement().getFirstChild(); entry_n != null; entry_n = entry_n.getNextSibling()) {
		if(entry_n.getNodeName().equals("Entry")) {
		    Element entry_e = (Element) entry_n;
		    String extension = entry_e.getAttribute("extension");
		    String command = MSMUtils.getValue(entry_e);
		    put(extension, command);
		    extension = null;
		    command = null;
		    entry_e = null;
		}
	    }
	}
    }

    private void save() {
	// Only save if there is something to save.
	if(size() > 0) {
	    StringBuffer assoc = new StringBuffer("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	    assoc.append("<!DOCTYPE Associations [\n");
	    assoc.append("  <!ELEMENT Associations (Entry*)>\n");
	    assoc.append("  <!ELEMENT Entry (#PCDATA)>\n");
	    assoc.append("  <!ATTLIST Entry\n");
	    assoc.append("            extension CDATA #REQUIRED>\n");
	    assoc.append("]>\n");
	    assoc.append("\n");
	    assoc.append("<Associations>\n");
	    for(Iterator keys = keySet().iterator(); keys.hasNext(); ) {
		String extension = (String) keys.next();
		String command = (String) get(extension);
		assoc.append("\t<Entry extension=\"");
		assoc.append(extension);
		assoc.append("\">");
		assoc.append(command);
		assoc.append("</Entry>\n");
	    }
	    assoc.append("</Associations>");

	    String assoc_str = assoc.toString();
	    assoc = null;
	    try {
		FileOutputStream fos = new FileOutputStream(ASSOC_FILE);
		OutputStreamWriter osw = new OutputStreamWriter(fos, Charset.forName("UTF-8"));
		int position = 0;
		int BLOCK_SIZE = 1024;
		// Write x block sized chunks to file.
		while(position + BLOCK_SIZE < assoc_str.length()) {
		    osw.write(assoc_str, position, BLOCK_SIZE);
		    position = position + BLOCK_SIZE;
		}
		// Write the remainder of the buffer.
		if(position < assoc_str.length()) {
		    osw.write(assoc_str, position, assoc_str.length() - position);
		}
		osw.flush();
		osw.close();
		osw = null;
		fos.close();
		fos = null;
		assoc_str = null;
	    }
	    catch (Exception error) {
		Gatherer.printStackTrace(error);
	    }
	}
    }

    public void setCommand(String extension, String command) {
	///ystem.err.println("Set command: " + extension + " = " + command);
	if(command == null || command.length() == 0) {
	    remove(extension);
	}
	else {
	    put(extension, command);
	}
    }
}



